@@ -1,5 +1,5 @@
 /*  lmpc -- the Little Movie Processing Centre
-    Copyright (C) 1994-98 Uwe Girlich
+    Copyright (C) 1994-99 Uwe Girlich
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
     Uwe Girlich
-    Erika-von-Brockdorff-Strasse 2
+    Dantestrasse 20
     04159 Leipzig
     Deutschland / Germany
     E-mail: Uwe.Girlich@itp.uni-leipzig */
@@ -59,9 +59,26 @@
 #ifdef ENABLE_DM2
   #include "udm2.h"
 #endif
+#ifdef ENABLE_DM3
+  #include "udm3.h"
+#else
+  int isDM3bin(char *filename) { return 0; }
+  int isDM3txt(char *filename) { return 0; }
+#endif
 
 /* prototypes */
-#ifndef HAVE_MEMMEM
+
+/* from the text parse code */
+#ifdef __cplusplus
+extern "C" {
+#endif
+extern FILE* yyin;
+extern int yyparse(void);
+#ifdef __cplusplus
+}
+#endif
+
+#ifndef HAVE_MEMMEM_PROTOTYPE
   extern void *memmem(const void *haystack, size_t haystack_len,
                       const void *needle, size_t needle_len);
 #endif           
@@ -96,6 +113,7 @@ void ActionQWDtxt2QWDtxt(char *qwdtxtfilename1, char *qwdtxtfilename2,
        opt_t *opt);
 
 void ActionInfoDM2(char *filename, opt_t *opt);
+void ActionInplaceDM2(char *filename, opt_t *opt);
 void ActionDM2bin2DM2bin(char *dm2binfilename1, char *dm2binfilename2,
        opt_t *opt);
 void ActionDM2bin2DM2txt(char *dm2binfilename, char *dm2txtfilename,
@@ -105,6 +123,17 @@ void ActionDM2txt2DM2bin(char *dm2txtfilename, char *dm2binfilename,
 void ActionDM2txt2DM2txt(char *dm2txtfilename1, char *dm2txtfilename2,
        opt_t *opt);
 
+void ActionInfoDM3(char *filename, opt_t *opt);
+void ActionInplaceDM3(char *filename, opt_t *opt);
+void ActionDM3bin2DM3bin(char *dm3binfilename1, char *dm3binfilename2,
+       opt_t *opt);
+void ActionDM3bin2DM3txt(char *dm3binfilename, char *dm3txtfilename,
+       opt_t *opt);
+void ActionDM3txt2DM3bin(char *dm3txtfilename, char *dm3binfilename,
+       opt_t *opt);
+void ActionDM3txt2DM3txt(char *dm3txtfilename1, char *dm3txtfilename2,
+       opt_t *opt);
+
 /* special routines */
 char *opt_string(opt_t *opt);
 
@@ -252,7 +281,8 @@ int main(int argc, char **argv)
       case 'g':
         if (opt.option&opDefineGame) syserror(MANYARGS,c_str);
         opt.option |= opDefineGame;
-        for(optargp=optarg;*optargp;optargp++) *optargp = toupper(*optargp);
+        for(optargp=optarg;*optargp;optargp++)
+          *optargp = (char) toupper((int)*optargp);
         if ((!strcmp(optarg,"DOOM")) || (!strcmp(optarg,"D"))) {
           opt.game=(DOOM_old|DOOM_new);
           break;
@@ -293,6 +323,12 @@ int main(int argc, char **argv)
           opt.game=QUAKE2;
           break;
         }
+#ifdef ENABLE_DM3
+        if ((!strcmp(optarg,"QUAKE3")) || (!strcmp(optarg,"A"))) {
+          opt.game=QUAKE3;
+          break;
+        }
+#endif
         syserror(INVOPT, optarg);
       break;
       case 'm':
@@ -318,7 +354,7 @@ int main(int argc, char **argv)
   /* without options means acInfo or acInplace */
   if (!opt.action) {
     if (opt.option&opFixMulti) {
-      opt.action |= acInplaceDEM;
+      opt.action |= acInplace;
     }
     else {
       opt.action |= acInfo;
@@ -336,7 +372,7 @@ int main(int argc, char **argv)
   /* --game only with info, --to-txt and convert */
   /* --game can be everywhere (at least with DEM), this prevents stdin reading */
   if ((opt.option&opDefineGame) &&
-      (opt.action&(~(acInfo|ac2TXT|ac2BIN|acInplaceDEM|acConvert)))) 
+      (opt.action&(~(acInfo|ac2TXT|ac2BIN|acInplace|acConvert)))) 
     syserror(ILLOPT, "--game");
 
   /* convert MS-DOS filenames to lower case, this may be deleted soon */
@@ -369,6 +405,12 @@ int main(int argc, char **argv)
             ActionInfoDM2(argv[argvf_base+i], &opt);
             continue;
           }
+#ifdef ENABLE_DM3          
+          if (opt.game==QUAKE3) {
+            ActionInfoDM3(argv[argvf_base+i], &opt);
+            continue;
+          }
+#endif          
           /* there is no LMP type check here */
           ActionInfoLMP(argv[argvf_base+i], &opt);
           continue;
@@ -389,6 +431,10 @@ int main(int argc, char **argv)
           ActionInfoDM2(argv[argvf_base+i], &opt);
           continue;
         }
+        if (isDM3bin(argv[argvf_base+i]) || isDM3txt(argv[argvf_base+i])) {
+          ActionInfoDM3(argv[argvf_base+i], &opt);
+          continue;
+        }
         ActionInfoLMP(argv[argvf_base+i], &opt);
         continue; /* in fact sense-less */
       }
@@ -403,12 +449,12 @@ int main(int argc, char **argv)
           break;
         }
         if (opt.game==QUAKE) {
-          if (isDEMtxt(argv[argvf_base])) {
-            ActionDEMtxt2DEMtxt(argv[argvf_base], argv[argvf_base+1], &opt);
+          if (isDEMbin(argv[argvf_base])) {
+            ActionDEMbin2DEMtxt(argv[argvf_base], argv[argvf_base+1], &opt);
             break;
           } 
           else {
-            ActionDEMbin2DEMtxt(argv[argvf_base], argv[argvf_base+1], &opt);
+            ActionDEMtxt2DEMtxt(argv[argvf_base], argv[argvf_base+1], &opt);
             break;
           }     
         }
@@ -432,6 +478,18 @@ int main(int argc, char **argv)
             break;
           }
         }
+#ifdef ENABLE_DM3        
+        if (opt.game==QUAKE3) {
+          if (isDM3txt(argv[argvf_base])) {
+            ActionDM3txt2DM3txt(argv[argvf_base], argv[argvf_base+1], &opt);
+            break;
+          }
+          else {
+            ActionDM3bin2DM3txt(argv[argvf_base], argv[argvf_base+1], &opt);
+            break;
+          }
+        }
+#endif        
         /* there is no LMP type check here */
         ActionLMP2LS(argv[argvf_base], argv[argvf_base+1], &opt);
         break;
@@ -464,6 +522,14 @@ int main(int argc, char **argv)
         ActionDM2bin2DM2txt(argv[argvf_base], argv[argvf_base+1], &opt);
         break;
       }
+      if (isDM3txt(argv[argvf_base])) {  
+        ActionDM3txt2DM3txt(argv[argvf_base], argv[argvf_base+1], &opt);
+        break;
+      }
+      if (isDM3bin(argv[argvf_base])) {  
+        ActionDM3bin2DM3txt(argv[argvf_base], argv[argvf_base+1], &opt);
+        break;
+      }
       ActionLMP2LS(argv[argvf_base], argv[argvf_base+1], &opt);
       /* break; in fact sense-less */
     break;
@@ -506,6 +572,20 @@ int main(int argc, char **argv)
         ActionDM2bin2DM2bin(argv[argvf_base], argv[argvf_base+1], &opt);
         break;
       }
+#ifdef ENABLE_DM3      
+      if (((opt.option&opDefineGame) && opt.game==QUAKE3) ||
+          (isDM3txt(argv[argvf_base]))) {
+        if (argf!=2) syserror(MANYARGS, "command line");
+        ActionDM3txt2DM3bin(argv[argvf_base], argv[argvf_base+1], &opt);
+        break;
+      }
+      if (((opt.option&opDefineGame) && opt.game==QUAKE3) ||
+          (isDM3bin(argv[argvf_base]))) {
+        if (argf!=2) syserror(MANYARGS, "command line");
+        ActionDM3bin2DM3bin(argv[argvf_base], argv[argvf_base+1], &opt);
+        break;
+      }
+#endif      
       /* argf >=2 */
       for (i=0;i<argf-1;i++) { /* count the LS files */
         /*
@@ -538,11 +618,21 @@ int main(int argc, char **argv)
       else
         ActionConvert(argv[argvf_base], argv[argvf_base+1], &opt);
     break;
-    case acInplaceDEM:
+    case acInplace:
       if (argf<1) syserror(MISIFN, "command line");
-      for (i=0;i<argf;i++) { 
-        ActionInplaceDEM(argv[argvf_base+i], &opt);
-      }    
+      for (i=0;i<argf;i++) {
+        if (isDEMbin(argv[argvf_base+i])) {
+          ActionInplaceDEM(argv[argvf_base+i], &opt);
+        }
+        else {
+          if (isDM2bin(argv[argvf_base+i])) {
+            ActionInplaceDM2(argv[argvf_base+i], &opt);
+          }
+          else {
+            syserror(NOTSUP,"game type of %s",argv[argvf_base+i]);
+          }
+        }
+      }
     break;
   }
 
@@ -557,7 +647,7 @@ int main(int argc, char **argv)
 void syntax(void)
 {
   fprintf(stderr,"Little Movie Processing Centre\n");
-  fprintf(stderr,"LMPC (c) U. Girlich, 1994-98, Release %s %s (%s)\n", VERSION, DATE, COMMENT);
+  fprintf(stderr,"LMPC (c) U. Girlich, 1994-99, Release %s %s (%s)\n", VERSION, DATE, COMMENT);
   fprintf(stderr,"lmpc [option ...] filename [filename ...]\n");
   fprintf(stderr,"-h, --help                   display this help and exit.\n");
   fprintf(stderr,"-V, --version                output version information and exit.\n");
@@ -571,12 +661,17 @@ void syntax(void)
   fprintf(stderr,"-w, --add-WT SECONDS         adds SECONDS (int) WT game tics to the LMP file\n");
   fprintf(stderr,"-r, --remove-pause           removes the game tics between PS and PE.\n");
   fprintf(stderr,"-c, --clear-save             clears SG tics.\n");
-  fprintf(stderr,"-f, --fix-multi              fixes multi-player DEM files.\n");
+  fprintf(stderr,"-f, --fix-multi              fixes multi-level DEM and DM2 files.\n");
   fprintf(stderr,"-t, --tic FROM(,|:)TO        converts only a part of the LMP-file\n");
   fprintf(stderr,"                             (in connection with -s only).\n");
-  fprintf(stderr,"-g, --game GAME              forces game (GAME=DOOM,DOOM2,HERETIC,HEXEN,\n");
-  fprintf(stderr,"                             STRIFE,DUKE,REDNECK,QUAKE,QUAKEWORLD,QUAKE2)\n");
+  fprintf(stderr,"-g, --game GAME              forces game (GAME=DOOM,DOOM2,HERETIC,HEXEN,STRIFE\n");
+#ifdef ENABLE_DM3
+  fprintf(stderr,"                             DUKE,REDNECK,QUAKE,QUAKEWORLD,QUAKE2,QUAKE3)\n");
+  fprintf(stderr,"                             short: D2HXSNRQWqA (only with -(i|s|v|p|w|r|c)).\n");
+#else  
+  fprintf(stderr,"                             DUKE,REDNECK,QUAKE,QUAKEWORLD,QUAKE2)\n");
   fprintf(stderr,"                             short: D2HXSNRQWq (only with -(i|s|v|p|w|r|c)).\n");
+#endif  
   fprintf(stderr,"-m, --mark-step STEP         marks every STEPth tic in the LS file (-s only)\n");
   fprintf(stderr,"                             STEP=0 doesn't mark any game tic; default: STEP=1.\n");
   exit(1);
@@ -639,7 +734,7 @@ void ActionInfoLMP(char *filename, opt_t *opt)
   if (l.game&(HEXEN_old|HEXEN_new)) {
     fprintf(stderr," (");
     for (i=0;i<l.playernum;i++) {
-      fprintf(stderr,"%s", ClassName[l.class[i]]);
+      fprintf(stderr,"%s", ClassName[l.playerclass[i]]);
       if (l.playernum>1 && i<l.playernum-1) fprintf(stderr,",");
     }
     fprintf(stderr,")");
@@ -671,7 +766,7 @@ void ActionLMP2LS(char *lmpfilename, char *lsfilename, opt_t *opt)
   double time;
   char ts[200];
   char cs[200];
-  unsigned long tic, stic;
+  long tic, stic;
   unsigned char b;
   int wb[8], ab, cb, wn;
   int KeyOld, KeyNew; 
@@ -711,7 +806,7 @@ void ActionLMP2LS(char *lmpfilename, char *lsfilename, opt_t *opt)
     break;
     case HEXEN_old: 
     case HEXEN_new:
-                     for (p=0;p<l.playernum;p++) wb[p]=24+l.class[p]*4; 
+                     for (p=0;p<l.playernum;p++) wb[p]=24+l.playerclass[p]*4; 
                      ab = 10; 
                      cb = 8;
                      wn = 4;
@@ -742,13 +837,13 @@ void ActionLMP2LS(char *lmpfilename, char *lsfilename, opt_t *opt)
     for (p=0;p<l.playernum;p++) {
       sprintf(ss," Player%d:      %d        # %s player",p,l.num[p],ColorName[cb+l.num[p]]); WriteLine(s.file,ss);
       if (l.game&(HEXEN_old|HEXEN_new)) {
-        sprintf(ss," Class%d:       %d        # %s",p,l.class[p],ClassName[l.class[p]]); WriteLine(s.file,ss);
+        sprintf(ss," Class%d:       %d        # %s",p,l.playerclass[p],ClassName[l.playerclass[p]]); WriteLine(s.file,ss);
       }
     }
   }
   else {
     if (l.game&(HEXEN_old|HEXEN_new)) {
-      sprintf(ss," Class:        %d        # %s",l.class[0],ClassName[l.class[0]]); WriteLine(s.file,ss);
+      sprintf(ss," Class:        %d        # %s",l.playerclass[0],ClassName[l.playerclass[0]]); WriteLine(s.file,ss);
     }
   }
   if (l.game&DOOM_new || l.game&DOOM2 || l.game&STRIFE) {
@@ -1151,7 +1246,7 @@ void ActionLS2LMP(char *lsfilename, char *lmpfilename, int action)
 	l.mainplayer=0;
 	for (i=0;i<MAXPLAYER_GLOBAL;i++) { 
           l.num[i]=0; 
-          l.class[i]=0;
+          l.playerclass[i]=0;
         }
 	if (strcmp(gt,"HEADERSTART")) syntaxerror(linenumber,"HeaderStart expected");
 	state=HEADER;
@@ -1277,55 +1372,55 @@ void ActionLS2LMP(char *lsfilename, char *lmpfilename, int action)
 	    notdecoded=FALSE;
 	  }
 	  if (notdecoded && strcmp(token,"CLASS")==0) {
-	    l.class[0]=number;
+	    l.playerclass[0]=number;
 	    if (l.playernum<1) l.playernum=1;
 	    numberused=TRUE;
 	    notdecoded=FALSE;
 	  }
 	  if (notdecoded && strcmp(token,"CLASS0")==0) {
-	    l.class[0]=number;
+	    l.playerclass[0]=number;
 	    if (l.playernum<1) l.playernum=1;
 	    numberused=TRUE;
 	    notdecoded=FALSE;
 	  }
 	  if (notdecoded && strcmp(token,"CLASS1")==0) {
-	    l.class[1]=number;
+	    l.playerclass[1]=number;
 	    if (l.playernum<1) l.playernum=1;
 	    numberused=TRUE;
 	    notdecoded=FALSE;
 	  }
 	  if (notdecoded && strcmp(token,"CLASS2")==0) {
-	    l.class[2]=number;
+	    l.playerclass[2]=number;
 	    if (l.playernum<1) l.playernum=1;
 	    numberused=TRUE;
 	    notdecoded=FALSE;
 	  }
 	  if (notdecoded && strcmp(token,"CLASS3")==0) {
-	    l.class[3]=number;
+	    l.playerclass[3]=number;
 	    if (l.playernum<1) l.playernum=1;
 	    numberused=TRUE;
 	    notdecoded=FALSE;
 	  }
           if (notdecoded && strcmp(token,"CLASS4")==0) {
-            l.class[4]=number;
+            l.playerclass[4]=number;
             if (l.playernum<1) l.playernum=1;
             numberused=TRUE;
             notdecoded=FALSE;
           }
           if (notdecoded && strcmp(token,"CLASS5")==0) {
-            l.class[5]=number;
+            l.playerclass[5]=number;
             if (l.playernum<1) l.playernum=1;
             numberused=TRUE;
             notdecoded=FALSE;
           }
           if (notdecoded && strcmp(token,"CLASS6")==0) {
-            l.class[6]=number;
+            l.playerclass[6]=number;
             if (l.playernum<1) l.playernum=1;
             numberused=TRUE;
             notdecoded=FALSE;
           }
           if (notdecoded && strcmp(token,"CLASS7")==0) {
-            l.class[7]=number;
+            l.playerclass[7]=number;
             if (l.playernum<1) l.playernum=1;
             numberused=TRUE;
             notdecoded=FALSE;
@@ -1705,7 +1800,7 @@ void ActionConvert(char *infilename, char *outfilename, opt_t *opt)
   int NewFile, NewTics, NewHeader; 
   int o_game;
   char *gs, *o_version;
-  unsigned long wttics=0, tic, stic, ps=0;
+  long wttics=0, tic, stic, ps=0;
   int p;
   TIC_t t;
   TIC_t WT = {0,0,0,0,0,0};
@@ -1965,7 +2060,7 @@ void ActionDMObin2DMOtxt(char *dmofilename, char *dsfilename, opt_t *opt)
   CHU_t chu;
   unsigned long i;
   int j;
-  unsigned long tics;
+  long tics;
   TEXT_t s;
   DMO_TIC_t t;
   char buffer[10];
@@ -1974,7 +2069,7 @@ void ActionDMObin2DMOtxt(char *dmofilename, char *dsfilename, opt_t *opt)
   double time;
   char ts[200];
   char cs[200];
-  unsigned long tic;
+  long tic;
   char ss[1000];
   int weapon;
 
@@ -2243,8 +2338,8 @@ void ActionInplaceDEM(char *filename, opt_t *opt)
              bufferbase++) {
           /* Note the parameter order! 
              The Linux memmem(3) man page is wrong */
-          if ((bufferpos=memmem(bufferbase, bufferend - bufferbase,
-                                FixMultiString, FixMultiStringLength))!=NULL) {
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FixMultiString, FixMultiStringLength))!=NULL) {
             /*
             fprintf(stdout,"found: ");
             fprintf(stdout,"0x%02x ",*bufferpos); 
@@ -2360,8 +2455,6 @@ void ActionDEMtxt2DEMtxt(char *demtxtfilename1, char *demtxtfilename2,
   TEXT_t s1;
   TEXT_t s2;
   char ts[1000];
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s1,demtxtfilename1,"rb");
   TEXT_init(&s2,demtxtfilename2,"wb");
@@ -2378,7 +2471,7 @@ void ActionDEMtxt2DEMtxt(char *demtxtfilename1, char *demtxtfilename2,
   WriteLine(s2.file,"");
 
   yyin = s1.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s1);
   TEXT_done(&s2);
@@ -2389,8 +2482,6 @@ void ActionDEMtxt2DEMbin(char *demtxtfilename, char *dembinfilename,
 {
   TEXT_t s;
   DEM_t d;
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s,demtxtfilename,"rb");
   DEM_init(&d,dembinfilename,"wb");
@@ -2404,7 +2495,7 @@ void ActionDEMtxt2DEMbin(char *demtxtfilename, char *dembinfilename,
                    demtxtfilename, dembinfilename, opt_string(opt));
 
   yyin = s.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s);
   DEM_done(&d);
@@ -2538,8 +2629,6 @@ void ActionQWDtxt2QWDbin(char *qwdtxtfilename, char *qwdbinfilename,
 {
   TEXT_t s;
   QWD_t d;
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s,qwdtxtfilename,"rb");
   QWD_init(&d,qwdbinfilename,"wb");
@@ -2553,7 +2642,7 @@ void ActionQWDtxt2QWDbin(char *qwdtxtfilename, char *qwdbinfilename,
                    qwdtxtfilename, qwdbinfilename, opt_string(opt));
 
   yyin = s.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s);
   QWD_done(&d);
@@ -2565,8 +2654,6 @@ void ActionQWDtxt2QWDtxt(char *qwdtxtfilename1, char *qwdtxtfilename2,
   TEXT_t s1;
   TEXT_t s2;
   char ts[1000];
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s1,qwdtxtfilename1,"rb");
   TEXT_init(&s2,qwdtxtfilename2,"wb");
@@ -2583,7 +2670,7 @@ void ActionQWDtxt2QWDtxt(char *qwdtxtfilename1, char *qwdtxtfilename2,
   WriteLine(s2.file,"");
 
   yyin = s1.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s1);
   TEXT_done(&s2);
@@ -2646,6 +2733,192 @@ void ActionInfoDM2(char *filename, opt_t *opt)
 
 }
 
+void ActionInplaceDM2(char *filename, opt_t *opt)
+{
+  DM2_t d;
+  unsigned char buffer[DM2_MAX_MESSAGE_SIZE];
+  size_t blockbase;
+  size_t blocksize;
+  unsigned char *bufferend; 
+  int changed;
+  unsigned char *bufferbase;
+  unsigned char *bufferpos;
+  int lastwasbetweenblock = 0;
+
+  #define FM_reconnect		"\013reconnect\n"
+  #define FMS_reconnect		(strlen(FM_reconnect)+1)
+  #define FM_configstrings	"\013cmd configstrings "
+  #define FMS_configstrings	(strlen(FM_configstrings))
+  #define FM_baselines		"\013cmd baselines "
+  #define FMS_baselines		(strlen(FM_baselines))
+  #define FM_precache		"\013precache "
+  #define FMS_precache		(strlen(FM_precache))
+  char FM_reconnect_m[]="\012\002Server restarted\n\001";
+  #define FMS_reconnect_m	sizeof(FM_reconnect_m)
+  *(FM_reconnect_m + FMS_reconnect_m - 2) = 0x00;
+  *(FM_reconnect_m + FMS_reconnect_m - 1) = 0x08;
+  #define FM_record		"\013record "
+  #define FMS_record		(strlen(FM_record))
+  #define FM_download		"\020\377\377"
+  #define FMS_download		(strlen(FM_download)+1)
+
+  if (isDM2bin(filename)) {   
+    DM2_init(&d, filename, "r+b");
+  
+    fprintf(stderr,"%s (DM2 bin) -> %s (DM2 bin)%s\n", d.filename, d.filename,
+                   opt_string(opt));
+    while ((blockbase=ftell(d.file)) < d.filesize) {
+      /*
+      fprintf(stderr,"blockbase=0x%lx, ", (unsigned long)blockbase);
+      */
+      /* I don't check return-values because of the successful isDM2bin */
+      /* read the block size from file */
+      fread(buffer+0,1,4,d.file);
+      /* get the block size */
+      blocksize = ((buffer[0])    ) +
+                  ((buffer[1])<< 8) +
+                  ((buffer[2])<<16) +
+                  ((buffer[3])<<24);
+      /* don't confuse with the endblock */
+      if (blocksize != 0xFFFFFFFF) {
+        /* read in the block */
+        fread(buffer+4,1,blocksize,d.file);
+        /* set the bufferend pointer at the end */
+        bufferend = buffer + 4 + blocksize;
+      }
+      else {
+        bufferend = buffer + 4;
+      }
+      changed = 0;
+
+      if (opt->option&opFixMulti) {
+        for (bufferbase = buffer + 4;
+             bufferbase < bufferend;
+             bufferbase++) {
+          /* Note the parameter order!
+             The Linux memmem(3) man page is wrong */
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_reconnect, FMS_reconnect))!=NULL) {
+            memset(bufferpos,0x06,FMS_reconnect); 
+            changed++;
+            bufferbase+=FMS_reconnect;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_configstrings, FMS_configstrings))!=NULL) {
+            size_t l;
+            memset(bufferpos,0x06,l = strlen((char*)bufferpos) + 1);
+            changed++;
+            bufferbase+=l;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_baselines, FMS_baselines))!=NULL) {
+            size_t l;
+            memset(bufferpos,0x06,l = strlen((char*)bufferpos) + 1);
+            changed++;
+            bufferbase+=l;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_record, FMS_record))!=NULL) {
+            size_t l;
+            memset(bufferpos,0x06,l = strlen((char*)bufferpos) + 1);
+            changed++;
+            bufferbase+=l;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_download, FMS_download))!=NULL) {
+            size_t l;
+            memset(bufferpos,0x06,l = FMS_download);
+            changed++;
+            bufferbase+=l;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_precache, FMS_precache))!=NULL) {
+            size_t l;
+            memset(
+                   /* after the new string end */
+                   bufferpos + FMS_precache + 1,
+                   /* NOP bytes */
+                   0x06,
+                   /* strlen + 1 times */
+                   l = strlen((char*)bufferpos + FMS_precache + 1) + 1
+                  );
+            *(bufferpos + FMS_precache - 1) = '\n'; /* \n instead of ' ' */ 
+            *(bufferpos + FMS_precache    ) = '\0'; /* \0 instead of anything */
+            changed++;
+            bufferbase+=l;
+          }
+          if ((bufferpos=(unsigned char*)memmem(bufferbase, bufferend - bufferbase,
+                                                FM_reconnect_m, FMS_reconnect_m))!=NULL) {
+            /* replace reconnect (0x08) with nop (0x06) */
+            *(bufferpos + FMS_reconnect_m - 1) = 0x06;
+            changed++;
+            bufferbase+=FMS_reconnect_m;
+          }
+          if (!changed) break;
+        } /* end for */
+        /* buffer[0..3] is the length */
+        /* buffer[4] is the first message id */
+        /* if it is reconnect, chang it to nop */
+        if ((blocksize > 4) && (buffer[4] == 0x08)) {
+          buffer[4] = 0x06;
+          changed++;
+        }
+        /* buffer[5..8] is serverversion */
+        /* buffer[9..12] is key */
+        /* buffer[13] is isdemo */
+        /* if it is serverdata, set isdemo to, if it was 0 before */
+        if ((blocksize > 13) && (buffer[4] == 0x0C) && (buffer[13] == 0x00)) {
+          buffer[13] = 0x01;
+          changed++;
+        }
+        if (lastwasbetweenblock) {
+          /* move the whole block (size plus data) further by 4 */
+          memmove(buffer + 4, buffer, blocksize + 4);
+          /* fill the old size bytes with NOPs */
+          memset(buffer + 4, 0x06, 4);
+          /* increase the current blocksize by 4 to include the betweenblock */
+          blocksize += 4;
+          /* put the new increased blocksize at the start: lowest byte first */
+          buffer[0] = ( blocksize      ) & 0xFF;
+          buffer[1] = ( blocksize >>  8) & 0xFF;
+          buffer[2] = ( blocksize >> 16) & 0xFF;
+          buffer[3] = ( blocksize >> 24) & 0xFF;
+          /* move the position in the file down by 4 */
+          blockbase -= 4;
+          /* switch this state off */
+          lastwasbetweenblock = 0;
+          /* write this block out */
+          changed++; 
+        }
+        if (blocksize == 0) {
+          lastwasbetweenblock = 1;
+        }
+      } /* end opFixMulti */
+      
+      /* here may come some other change or repair code */
+      
+      if (changed) {
+        /* something has changed, let's write it back in the file */
+        /* move back to the blockbase */
+        fseek(d.file, blockbase, SEEK_SET);
+        if (blocksize == 0xFFFFFFFF) {
+        /* endblock: 4 bytes */
+          fwrite(buffer, 1, 4, d.file);
+        }
+        else {
+          /* normal block: 4 + blocksize bytes */
+          fwrite(buffer, 1, 4 + blocksize, d.file);
+        }
+      }
+      /* fprintf(stdout,"\n"); fflush(stdout); */
+    } 
+    DM2_done(&d);
+  }
+  else {
+    syswarning(WDM2,filename);
+  }
+}
+
 void ActionDM2bin2DM2bin(char *dm2binfilename1, char *dm2binfilename2,
        opt_t *opt)
 {
@@ -2706,8 +2979,12 @@ void ActionDM2bin2DM2txt(char *dm2binfilename, char *dm2txtfilename,
   TEXT_init(&s,dm2txtfilename,"wb");
   node_write_text_init(&s); /* for node_write_text */
   node_token_init(DM2_token); /* for token routines */
+  glob_opt = opt; /* for DM2_block_edit */
+  do_block_edit = DM2_block_edit;
+  do_block_output = DM2_block_write_text;
 
-  fprintf(stderr,"%s (DM2 bin) -> %s (DM2 txt)\n", d.filename, s.filename);
+  fprintf(stderr,"%s (DM2 bin) -> %s (DM2 txt)%s\n", 
+                 d.filename, s.filename, opt_string(opt));
 
   sprintf(ts,"// source: DM2 binary file %s", d.filename); WriteLine(s.file, ts);
   WriteLine(s.file,"");
@@ -2724,7 +3001,7 @@ void ActionDM2bin2DM2txt(char *dm2binfilename, char *dm2txtfilename,
       m.hint |= DM2_RECORD_SERVER;
     }
     n=DM2_bin_to_node(&m,o);
-    DM2_block_write_text(n);
+    do_block(n);
     node_delete(n);
     d.frame++;
   }
@@ -2739,8 +3016,6 @@ void ActionDM2txt2DM2bin(char *dm2txtfilename, char *dm2binfilename,
 {
   TEXT_t s;
   DM2_t d;
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s,dm2txtfilename,"rb");
   DM2_init(&d,dm2binfilename,"wb");
@@ -2754,7 +3029,7 @@ void ActionDM2txt2DM2bin(char *dm2txtfilename, char *dm2binfilename,
                    dm2txtfilename, dm2binfilename, opt_string(opt));
 
   yyin = s.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s);
   DM2_done(&d);
@@ -2766,8 +3041,6 @@ void ActionDM2txt2DM2txt(char *dm2txtfilename1, char *dm2txtfilename2,
   TEXT_t s1;
   TEXT_t s2;
   char ts[1000];
-  extern FILE* yyin;
-  extern int dem_parse(void);
 
   TEXT_init(&s1,dm2txtfilename1,"rb");
   TEXT_init(&s2,dm2txtfilename2,"wb");
@@ -2784,7 +3057,7 @@ void ActionDM2txt2DM2txt(char *dm2txtfilename1, char *dm2txtfilename2,
   WriteLine(s2.file,"");
 
   yyin = s1.file;
-  dem_parse();
+  yyparse();
 
   TEXT_done(&s1);
   TEXT_done(&s2);
@@ -2818,6 +3091,116 @@ void ActionDM2txt2DM2txt(char *dm2txtfilename1, char *dm2txtfilename2,
 
 
 
+/*---------------------------------------------------------------------------*/
+/* DM2 */
+
+
+#ifdef ENABLE_DM3
+
+
+void ActionInfoDM3(char *filename, opt_t *opt)
+{
+  unsigned long tics;
+
+  udm3_init();
+
+  /* I check here again for text or bin, text check is much faster */
+  if (isDM3txt(filename)) {
+    fprintf(stderr,"%s: Quake III Arena (DM3 txt)\n", filename);
+  }
+  else {
+    if ((tics=isDM3bin(filename))!=0) {
+      fprintf(stderr,"%s: Quake III Arena (DM3 bin), %li frames\n", filename, tics);
+    }
+    else {
+      syswarning(WDM3,filename);
+    }
+  }
+  udm3_done();
+}
+
+void ActionDM3bin2DM3bin(char *dm3binfilename1, char *dm3binfilename2,
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3bin2DM3txt(char *dm3binfilename, char *dm3txtfilename, 
+       opt_t *opt)
+{
+  DM3_t d;
+  TEXT_t s;
+  char ts[1000];
+  BB_t m;
+  node *n;
+  int o;
+
+  udm3_init();
+
+  DM3_init(&d,dm3binfilename,"rb");
+  TEXT_init(&s,dm3txtfilename,"wb");
+  node_write_text_init(&s); /* for node_write_text */
+  node_token_init(DM3_token); /* for token routines */
+  glob_opt = opt; /* for DM3_block_edit */
+  do_block_edit = DM3_block_edit;
+  do_block_output = DM3_block_write_text;
+
+  fprintf(stderr,"%s (DM3 bin) -> %s (DM3 txt)%s\n", 
+                 d.filename, s.filename, opt_string(opt));
+
+  sprintf(ts,"// source: DM3 binary file %s", d.filename); WriteLine(s.file, ts);
+  WriteLine(s.file,"");
+
+  if (opt->option & opMarkStep) o=0x01; else o=0x00;
+  
+  d.frame=0;
+  while ((unsigned long)ftell(d.file)<d.filesize) {
+    DM3_block_read_bin(&d, &m);
+    n=DM3_bin_to_node(&m,o);
+    do_block(n);
+    node_delete(n);
+    d.frame++;
+  }
+  DM3_done(&d);
+  TEXT_done(&s);
+
+  udm3_done(); 
+}
+
+void ActionDM3txt2DM3bin(char *dm3txtfilename, char *dm3binfilename,
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3txt2DM3txt(char *dm3txtfilename1, char *dm3txtfilename2, 
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+
+#else /* !ENABLE_DM3" */
+
+
+void ActionInfoDM3(char *filename, opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3bin2DM3bin(char *dm3binfilename1, char *dm2binfilename2,
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3bin2DM3txt(char *dm3binfilename, char *dm3txtfilename, 
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3txt2DM3bin(char *dm3txtfilename, char *dm3binfilename,
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+void ActionDM3txt2DM3txt(char *dm3txtfilename1, char *dm3txtfilename2, 
+       opt_t *opt)
+{ syserror(NOTCOM,"DM3"); }
+
+
+#endif /* ENABLE_DM3 */
+
+
+
 /*****************************************************************************/
 /* special general routines **************************************************/
 /*****************************************************************************/
